﻿using System;
using System.Runtime.InteropServices;

/*
#include "stdafx.h"

void getCpuId(int bits[])
{
    _asm
    {
        xor ebx, ebx;
        xor ecx, ecx;
        xor edx, edx;
        mov eax, 0;

        cpuid;

        mov edi, bits;

        mov dword ptr [edi + 0], eax;
        mov dword ptr [edi + 4], ebx;
        mov dword ptr [edi + 8], ecx;
        mov dword ptr [edi + 12], edx;
    }
}

int _tmain(int argc, _TCHAR* argv[])
{
  int bits[4];

  getCpuId(bits); 

  printf("%d, %d, %d, %d\n", bits[0], bits[1], bits[2], bits[3]);
  printf("%x, %x, %x, %x\n", bits[0], bits[1], bits[2], bits[3]);

  return 0;
}
*/

namespace _3_DelegateWithASM
{
    // techdays 미니 토요 세미나 시리즈 - 한방에 따라잡는 C# 2.0 ~ 5.0
    // C#2.pptx

    // C++ 의 inline asm 사용을 .NET 으로 포팅하는 방법
    // http://www.sysnet.pe.kr/2/0/1267

    // 관련 글 - Visual Studio에서 직접 컴파일하는 IL 언어 확장 도구 - IL Support
    // http://www.sysnet.pe.kr/2/0/1697

    class Program
    {
        private readonly static byte[] x86CpuIdBytes =
        {
            0x55,
            0x8B, 0xEC,
            0x53,
            0x57,

            0x33, 0xDB,
            0x33, 0xC9,
            0x33, 0xD2,
            0xB8, 0x00, 0x00, 0x00, 0x00,

            0x0F, 0xA2, // cpuid

            0x8B, 0x7D, 0x08,

            0x89, 0x07,
            0x89, 0x5F, 0x04,
            0x89, 0x4F, 0x08,
            0x89, 0x57, 0x0C,

            0x5F,
            0x5B,

            0x5D,
            0xC3, // ret
        };

        [Flags]
        private enum AllocationType : uint
        {
            COMMIT = 0x1000,
            RESERVE = 0x2000,
            RESET = 0x80000,
            LARGE_PAGES = 0x20000000,
            PHYSICAL = 0x400000,
            TOP_DOWN = 0x100000,
            WRITE_WATCH = 0x200000
        }

        [Flags]
        public enum MemoryProtection : uint
        {
            EXECUTE = 0x10,
            EXECUTE_READ = 0x20,
            EXECUTE_READWRITE = 0x40,
            EXECUTE_WRITECOPY = 0x80,
            NOACCESS = 0x01,
            READONLY = 0x02,
            READWRITE = 0x04,
            WRITECOPY = 0x08,
            GUARD_Modifierflag = 0x100,
            NOCACHE_Modifierflag = 0x200,
            WRITECOMBINE_Modifierflag = 0x400
        }

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern IntPtr VirtualAlloc(IntPtr lpAddress, UIntPtr dwSize, AllocationType flAllocationType,
            MemoryProtection flProtect);

        [DllImport("kernel32")]
        private static extern bool VirtualFree(IntPtr lpAddress, UInt32 dwSize, UInt32 dwFreeType);

        private static IntPtr _codePointer;
        
        [UnmanagedFunctionPointerAttribute(CallingConvention.Cdecl)]
        private delegate void CpuIDDelegate(byte[] buffer);

        static CpuIDDelegate _cpuIdDelg;

        static void Main(string[] args)
        {
            try
            {
                byte[] codeBytes = x86CpuIdBytes;

                if (IntPtr.Size != 4)
                {
                    throw new NotSupportedException();
                }

                // 기계어를 담을 메모리 할당 후,
                _codePointer = VirtualAlloc(IntPtr.Zero, new UIntPtr((uint)codeBytes.Length),
                    AllocationType.COMMIT | AllocationType.RESERVE,
                    MemoryProtection.EXECUTE_READWRITE
                );

                Marshal.Copy(codeBytes, 0, _codePointer, codeBytes.Length);

                // 해당 주소에 대해 delegate를 가리키도록 설정
                _cpuIdDelg = (CpuIDDelegate)Marshal.GetDelegateForFunctionPointer(
                        _codePointer, typeof(CpuIDDelegate));

                if (_cpuIdDelg == null)
                {
                    throw new ObjectDisposedException("GetCpu");
                }

                GCHandle handle = default(GCHandle);

                try
                {
                    handle = GCHandle.Alloc(x86CpuIdBytes, GCHandleType.Pinned);

                    byte[] cpuIdBytes = new byte[4 * 4];

                    // 동적으로 메모리에 구성한 함수를 delegate를 사용해 호출
                    _cpuIdDelg(cpuIdBytes);
                    Console.WriteLine("Cpu Id: " + BitConverter.ToString(cpuIdBytes));
                }
                finally
                {
                    if (handle != default(GCHandle))
                    {
                        handle.Free();
                    }
                }
            }
            finally
            {
                if (_codePointer != IntPtr.Zero)
                {
                    VirtualFree(_codePointer, 0, 0x8000);
                    _codePointer = IntPtr.Zero;
                }
            }
        }
    }
}

/* cpuid
Cpu Id: 16-00-00-00-47-65-6E-75-6E-74-65-6C-69-6E-65-49
*/
