﻿using Microsoft.Win32.SafeHandles;
using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Security;
using System.Threading;

namespace NativeWithSpan
{
    // C# 7.2 - Span<T>
    // http://www.sysnet.pe.kr/2/0/11534

    class Program
    {
        static unsafe void Main(string[] args)
        {
            Thread t = new Thread(checkGCFunc);
            t.IsBackground = true;
            t.Start();


            // while (true)
            {
                /*
                0 : 0
                1 : 2226
                2 : 2149
                3 : 2308
                4 : 2279
                5 : 2289
                6
                */
                //int[] buf = new int[1024];
                //{
                //    for (int i = 0; i < buf.Length; i++)
                //    {
                //        buf[i] = i;
                //    }
                //}
            }

            while (true)
            {
                /*
                0 : 0
                1 : 0
                2 : 0
                3 : 0
                */
                using (NativeMemory<int> buf = new NativeMemory<int>(1024))
                {
                    Span<int> viewBuf = buf.GetView();
                    for (int i = 0; i < viewBuf.Length; i++)
                    {
                        viewBuf[i] = i;
                    }
                }
            }

            // viewBuf[1025] = 100; // 런타임 시 예외 발생 - System.IndexOutOfRangeException 

            /*
            try
            {
                IntPtr ptr = Marshal.AllocHGlobal(1);
                int* pBuf = (int*)ptr.ToPointer();

                *(pBuf + 20000) = 5; // 운이 좋으면 발생하지 않음.
                                        // Fatal error. System.AccessViolationException: Attempted to read or write protected memory. This is often an indication that other memory is corrupt.
                                        //        at NativeWithSpan.Program.Main(System.String[])
            }
            catch (Exception e)
            {
                // 예외에 잡히지 않고 비정상 종료.
                Console.WriteLine(e.ToString());
            }
            */
        }

        private static void checkGCFunc(object obj)
        {
            int old = 0;
            int checkCount = 0;

            while (true)
            {
                int count = 0;

                for (int i = 0; i < GC.MaxGeneration; i++)
                {
                    count += GC.CollectionCount(i);
                }

                Console.WriteLine($"{checkCount++} : {(count - old)}");
                old = count;

                Thread.Sleep(5000);
            }
        }

        public void Sample()
        {
            {
                string input = "123,456";
                int commaPos = input.IndexOf(',');
                int number123 = int.Parse(input.Substring(0, commaPos)); // "123" 신규 문자열 할당
                int number456 = int.Parse(input.Substring(commaPos + 1)); // "456" 신규 문자열 할당
            }

            {
                string input = "123,456";
                ReadOnlySpan<char> inputSpan = input.AsSpan();
                int commaPos = input.IndexOf(',');
                int number123 = int.Parse(inputSpan.Slice(0, commaPos));  // 할당 없이 "123" 영역을 가리키는 Span
                int number456 = int.Parse(inputSpan.Slice(commaPos + 1)); // 할당 없이 "456" 영역을 가리키는 Span

                Console.WriteLine($"{number123}, {number456}");
            }
        }
    }


    public unsafe ref struct NativeMemory<T> where T : unmanaged
    {
        int _size;
        IntPtr _ptr;

        public NativeMemory(int size)
        {
            _size = size;

            long lSize = _size;
            lSize *= sizeof(T);
                
            IntPtr bufSize = new IntPtr(lSize);
            _ptr = Marshal.AllocHGlobal(bufSize);
        }

        public Span<T> GetView()
        {
            return new Span<T>(_ptr.ToPointer(), _size);
        }

        // C# 8.0에서만 using과 함께 사용 가능
        public void Dispose()
        {
            if (_ptr == IntPtr.Zero)
            {
                return;
            }

            Marshal.FreeHGlobal(_ptr);
            _ptr = IntPtr.Zero;
        }
    }
}
